unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  math, FPImage;

const
  PCount = 1000; // Number of points in the light's outer circle
  LightStep = 0.9;
  LightDist = 200;

type

  { TForm1 }

  TForm1 = class(TForm)
    Image1: TImage;
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure DrawScreen;
    procedure Image1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Image1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer
      );
    procedure Timer1Timer(Sender: TObject);
  private
    back, front, lightPic: TPicture;
    mx, my: integer;
    map: array of array of boolean;
    p: array[0..PCount-1] of TPoint;
    fade: array of array of single;
    lightColor: TFPColor;
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

// Make value range 0..1 smooth like sine-wave
function Smoothen(n: single): single;
begin
  if n<0 then n:=0
  else if n>1 then n:=1;
  result:=0.5-cos(n*pi)*0.5;
end;

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
var i, j: integer;
    shapes: TPicture;
begin
  shapes:=TPicture.Create;
  shapes.LoadFromFile('shapes.png');
  setlength(map, shapes.Width, shapes.Height);
  with shapes.Bitmap.Canvas do begin
    for j:=0 to shapes.Height-1 do
      for i:=0 to shapes.Width-1 do
        map[i, j]:=(colors[i, j].red>30000);
  end;
  shapes.Free;

  back:=TPicture.Create;
  back.LoadFromFile('back.png');
  front:=TPicture.Create;
  front.LoadFromFile('front.png');
  lightPic:=TPicture.Create;
  lightPic.Bitmap.SetSize(front.Width, front.Height);

  // Generate quarter of circular fade map - bright center 1, fades outwards to 0
  setlength(fade, LightDist+1, LightDist+1);
  for j:=0 to LightDist do
    for i:=0 to LightDist do begin
      fade[i, j]:=1-smoothen(hypot(i, j)/LightDist);
    end;

  lightColor.red:=  140 *256;
  lightColor.green:=140 *256;
  lightColor.blue:= 20  *256;
  lightColor.alpha:=255 *256;

  image1.Picture.Bitmap.SetSize(front.Width, front.Height);
  image1.Align:=alClient;
  DrawScreen;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  back.Free;
  front.Free;
  lightPic.Free;
end;

procedure TForm1.DrawScreen;
var i, j, n, numSteps: integer;
    px, py, dx, dy, rnd, f: single;
    c, cb, cl: TFPColor;
begin
  if front=nil then exit;

  // Find light collisions
  numSteps:=round(LightDist/LightStep);
  rnd:=random;
  for i:=0 to PCount-1 do begin
    px:=mx/2;
    py:=my/2;
    dx:=cos(rnd+i/PCount*PI*2)*LightStep;
    dy:=sin(rnd+i/PCount*PI*2)*LightStep;
    for n:=1 to numSteps do begin
      px:=px+dx;
      py:=py+dy;
      if (px<0) or (py<0) or (px>high(map)) or (py>high(map[0])) or
         map[round(px), round(py)] then break;
    end;
    p[i].x:=round(px)*2;
    p[i].y:=round(py)*2;
  end;

  with lightPic.Bitmap.Canvas do begin
    brush.Color:=clBlack;
    Clear;
    brush.Color:=FPColorToTColor(lightColor);
    Polygon(p);
  end;

  c.alpha:=65535;
  with image1.Picture.Bitmap.Canvas do begin
    image1.Picture.Bitmap.BeginUpdate(true);
    //Draw(0, 0, back.Bitmap);
    //Draw(0, 0, front.Bitmap);
    //pen.Color:=clYellow;
    //brush.Color:=RGBToColor(230,230,0);
    //Polygon(p);
    //for i:=0 to PCount-1 do begin
      //MoveTo(mx, my);
      //LineTo(p[i]);
      //pixels[p[i].x, p[i].y]:=clRed;
    //end;

    //Draw(0, 0, lightPic.Bitmap);
    for j:=0 to lightPic.Height-1 do
      for i:=0 to lightPic.Width-1 do begin
        cb:=back.Bitmap.Canvas.Colors[i, j];
        cl:=lightPic.Bitmap.Canvas.Colors[i, j];
        if cl.red > 0 then begin
          //Colors[i, j]:=TColorToFPColor(clBlack);
          f:=fade[min(LightDist, abs(i-mx)), min(LightDist, abs(j-my))];
          c.red:=min(65535, cb.red+round(cl.red*f));
          c.green:=min(65535, cb.green+round(cl.green*f));
          c.blue:=min(65535, cb.blue+round(cl.blue*f));
          Colors[i, j]:=c;
        end else Colors[i, j]:=cb;
      end;

    Draw(0, 0, front.Bitmap);
    pixels[mx, my]:=clYellow;
    image1.Picture.Bitmap.EndUpdate();
  end;
end;

procedure TForm1.Image1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  lightPic.SaveToFile('_lightPic_temp.png');
end;

procedure TForm1.Image1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var baseRatio, ratio, scale: single;
begin
  baseRatio:=front.Width/front.Height;
  ratio:=image1.Width/image1.Height;

  if ratio>baseRatio then begin
    // Window is wider than image (empty strides at sides)
    scale:=image1.Height/front.Height;
    mx:=round((X-(image1.Width-image1.Height*baseRatio)/2)/scale);
    my:=round(Y/scale);
  end else begin
    // (empty strides top and bottom)
    scale:=image1.Width/front.Width;
    mx:=round(X/scale);
    my:=round((Y-(image1.Height-image1.width/baseRatio)/2)/scale);
  end;
  //DrawScreen;

  // Debug values (don't guess too much how things work, test them)
  //caption:=format('%.2f : %.2f : %.2f',[ratio, baseRatio, scale]);
  // Debug cursor position
  //caption:=format('%d : %d',[mx, my]);
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  DrawScreen;
end;

end.

